@@ -15,7 +15,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
  */
 
 #ifndef AGENT_H
@@ -62,7 +62,6 @@ struct
   int quiet;           /* Be as quiet as possible */
   int dry_run;         /* Don't change any persistent data */
   int batch;           /* Batch mode */
-  const char *homedir; /* Configuration directory name */
 
   /* True if we handle sigusr2.  */
   int sigusr2_enabled;
@@ -123,7 +122,7 @@ struct
 
   /* If set, a passphrase history will be written and checked at each
      passphrase change.  */
-  int enable_passhrase_history;
+  int enable_passphrase_history;
 
   int running_detached; /* We are running detached from the tty. */
 
@@ -155,12 +154,12 @@ struct
   int keep_tty;      /* Don't switch the TTY (for pinentry) on request */
   int keep_display;  /* Don't switch the DISPLAY (for pinentry) on request */
 
-  /* This global options indicates the use of an extra socket. Note
+  /* This global option indicates the use of an extra socket. Note
      that we use a hack for cleanup handling in gpg-agent.c: If the
      value is less than 2 the name has not yet been malloced. */
   int extra_socket;
 
-  /* This global options indicates the use of an extra socket for web
+  /* This global option indicates the use of an extra socket for web
      browsers. Note that we use a hack for cleanup handling in
      gpg-agent.c: If the value is less than 2 the name has not yet
      been malloced. */
@@ -169,7 +168,6 @@ struct
 
 
 /* Bit values for the --debug option.  */
-#define DBG_COMMAND_VALUE 1	/* debug commands i/o */
 #define DBG_MPI_VALUE	  2	/* debug mpi details */
 #define DBG_CRYPTO_VALUE  4	/* debug low level crypto */
 #define DBG_MEMORY_VALUE  32	/* debug memory allocation stuff */
@@ -179,7 +177,6 @@ struct
 #define DBG_IPC_VALUE     1024  /* Enable Assuan debugging.  */
 
 /* Test macros for the debug option.  */
-#define DBG_COMMAND (opt.debug & DBG_COMMAND_VALUE)
 #define DBG_CRYPTO  (opt.debug & DBG_CRYPTO_VALUE)
 #define DBG_MEMORY  (opt.debug & DBG_MEMORY_VALUE)
 #define DBG_CACHE   (opt.debug & DBG_CACHE_VALUE)
@@ -342,6 +339,7 @@ void agent_set_progress_cb (void (*cb)(ctrl_t ctrl, const char *what,
 gpg_error_t agent_copy_startup_env (ctrl_t ctrl);
 const char *get_agent_socket_name (void);
 const char *get_agent_ssh_socket_name (void);
+int get_agent_active_connection_count (void);
 #ifdef HAVE_W32_SYSTEM
 void *get_agent_scd_notify_event (void);
 #endif
@@ -349,7 +347,8 @@ void agent_sighup_action (void);
 int map_pk_openpgp_to_gcry (int openpgp_algo);
 
 /*-- command.c --*/
-gpg_error_t agent_inq_pinentry_launched (ctrl_t ctrl, unsigned long pid);
+gpg_error_t agent_inq_pinentry_launched (ctrl_t ctrl, unsigned long pid,
+                                         const char *extra);
 gpg_error_t agent_write_status (ctrl_t ctrl, const char *keyword, ...)
      GPGRT_ATTR_SENTINEL(0);
 gpg_error_t agent_print_status (ctrl_t ctrl, const char *keyword,
@@ -490,6 +489,9 @@ gpg_error_t s2k_hash_passphrase (const char *passphrase, int hashalgo,
                                  const unsigned char *s2ksalt,
                                  unsigned int s2kcount,
                                  unsigned char *key, size_t keylen);
+gpg_error_t agent_write_shadow_key (const unsigned char *grip,
+                                    const char *serialno, const char *keyid,
+                                    const unsigned char *pkbuf, int force);
 
 
 /*-- trustlist.c --*/
