@@ -15,7 +15,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <config.h>
@@ -25,7 +25,6 @@
 #include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <glob.h>
 #include <gpg-error.h>
 #include <stdarg.h>
 #include <stdlib.h>
@@ -36,6 +35,7 @@
 #include <unistd.h>
 
 #if HAVE_LIBREADLINE
+#define GNUPG_LIBREADLINE_H_INCLUDED
 #include <readline/readline.h>
 #include <readline/history.h>
 #endif
@@ -48,6 +48,20 @@
 #include "ffi.h"
 #include "ffi-private.h"
 
+/* For use in nice error messages.  */
+static const char *
+ordinal_suffix (int n)
+{
+  switch (n)
+    {
+    case 1: return "st";
+    case 2: return "nd";
+    case 3: return "rd";
+    default: return "th";
+    }
+  assert (! "reached");
+}
+
 
 
 int
@@ -237,7 +251,9 @@ do_setenv (scheme *sc, pointer args)
   FFI_ARG_OR_RETURN (sc, char *, value, string, args);
   FFI_ARG_OR_RETURN (sc, int, overwrite, bool, args);
   FFI_ARGS_DONE_OR_RETURN (sc, args);
-  FFI_RETURN_ERR (sc, gnupg_setenv (name, value, overwrite));
+  if (gnupg_setenv (name, value, overwrite))
+    FFI_RETURN_ERR (sc, gpg_error_from_syserror ());
+  FFI_RETURN (sc);
 }
 
 static pointer
@@ -309,6 +325,21 @@ do_close (scheme *sc, pointer args)
   FFI_RETURN_ERR (sc, close (fd) == 0 ? 0 : gpg_error_from_syserror ());
 }
 
+static pointer
+do_seek (scheme *sc, pointer args)
+{
+  FFI_PROLOG ();
+  int fd;
+  off_t offset;
+  int whence;
+  FFI_ARG_OR_RETURN (sc, int, fd, number, args);
+  FFI_ARG_OR_RETURN (sc, off_t, offset, number, args);
+  FFI_ARG_OR_RETURN (sc, int, whence, number, args);
+  FFI_ARGS_DONE_OR_RETURN (sc, args);
+  FFI_RETURN_ERR (sc, lseek (fd, offset, whence) == (off_t) -1
+                  ? gpg_error_from_syserror () : 0);
+}
+
 static pointer
 do_mkdtemp (scheme *sc, pointer args)
 {
@@ -460,6 +491,83 @@ do_rmdir (scheme *sc, pointer args)
   FFI_RETURN (sc);
 }
 
+static pointer
+do_get_isotime (scheme *sc, pointer args)
+{
+  FFI_PROLOG ();
+  gnupg_isotime_t timebuf;
+  FFI_ARGS_DONE_OR_RETURN (sc, args);
+  gnupg_get_isotime (timebuf);
+  FFI_RETURN_STRING (sc, timebuf);
+}
+
+static pointer
+do_getpid (scheme *sc, pointer args)
+{
+  FFI_PROLOG ();
+  FFI_ARGS_DONE_OR_RETURN (sc, args);
+  FFI_RETURN_INT (sc, getpid ());
+}
+
+static pointer
+do_srandom (scheme *sc, pointer args)
+{
+  FFI_PROLOG ();
+  int seed;
+  FFI_ARG_OR_RETURN (sc, int, seed, number, args);
+  FFI_ARGS_DONE_OR_RETURN (sc, args);
+  srand (seed);
+  FFI_RETURN (sc);
+}
+
+static int
+random_scaled (int scale)
+{
+  int v;
+#ifdef HAVE_RAND
+  v = rand ();
+#else
+  v = random ();
+#endif
+
+#ifndef RAND_MAX   /* for SunOS */
+#define RAND_MAX 32767
+#endif
+
+  return ((int) (1 + (int) ((float) scale * v / (RAND_MAX + 1.0))) - 1);
+}
+
+static pointer
+do_random (scheme *sc, pointer args)
+{
+  FFI_PROLOG ();
+  int scale;
+  FFI_ARG_OR_RETURN (sc, int, scale, number, args);
+  FFI_ARGS_DONE_OR_RETURN (sc, args);
+  FFI_RETURN_INT (sc, random_scaled (scale));
+}
+
+static pointer
+do_make_random_string (scheme *sc, pointer args)
+{
+  FFI_PROLOG ();
+  int size;
+  pointer chunk;
+  char *p;
+  FFI_ARG_OR_RETURN (sc, int, size, number, args);
+  FFI_ARGS_DONE_OR_RETURN (sc, args);
+  if (size < 0)
+    return ffi_sprintf (sc, "size must be positive");
+
+  chunk = sc->vptr->mk_counted_string (sc, NULL, size);
+  if (sc->no_memory)
+    FFI_RETURN_ERR (sc, ENOMEM);
+
+  for (p = sc->vptr->string_value (chunk); size; p++, size--)
+    *p = (char) random_scaled (256);
+  FFI_RETURN_POINTER (sc, chunk);
+}
+
 
 
 /* estream functions.  */
@@ -644,7 +752,7 @@ do_spawn_process (scheme *sc, pointer args)
     }
 
   err = gnupg_spawn_process (argv[0], (const char **) &argv[1],
-                             GPG_ERR_SOURCE_DEFAULT,
+                             NULL,
                              NULL,
                              flags,
                              &infp, &outfp, &errfp, &pid);
@@ -748,17 +856,19 @@ do_wait_processes (scheme *sc, pointer args)
 
   err = ffi_list2argv (sc, list_names, &names, &count);
   if (err == gpg_error (GPG_ERR_INV_VALUE))
-    return ffi_sprintf (sc, "%luth element of first argument is "
+    return ffi_sprintf (sc, "%lu%s element of first argument is "
                         "neither string nor symbol",
-                        (unsigned long) count);
+                        (unsigned long) count,
+                        ordinal_suffix ((int) count));
   if (err)
     FFI_RETURN_ERR (sc, err);
 
   err = ffi_list2intv (sc, list_pids, (int **) &pids, &count);
   if (err == gpg_error (GPG_ERR_INV_VALUE))
-    return ffi_sprintf (sc, "%luth element of second argument is "
-                        "neither string nor symbol",
-                        (unsigned long) count);
+    return ffi_sprintf (sc, "%lu%s element of second argument is "
+                        "not a number",
+                        (unsigned long) count,
+                        ordinal_suffix ((int) count));
   if (err)
     FFI_RETURN_ERR (sc, err);
 
@@ -921,30 +1031,42 @@ do_splice (scheme *sc, pointer args)
 {
   FFI_PROLOG ();
   int source;
-  int sink;
-  ssize_t len = -1;
   char buffer[1024];
   ssize_t bytes_read;
+  pointer sinks, sink;
   FFI_ARG_OR_RETURN (sc, int, source, number, args);
-  FFI_ARG_OR_RETURN (sc, int, sink, number, args);
-  if (args != sc->NIL)
-    FFI_ARG_OR_RETURN (sc, ssize_t, len, number, args);
-  FFI_ARGS_DONE_OR_RETURN (sc, args);
-  while (len == -1 || len > 0)
-    {
-      size_t want = sizeof buffer;
-      if (len > 0 && (ssize_t) want > len)
-        want = (size_t) len;
+  sinks = args;
+  if (sinks == sc->NIL)
+    return ffi_sprintf (sc, "need at least one sink");
+  for (sink = sinks; sink != sc->NIL; sink = pair_cdr (sink), ffi_arg_index++)
+    if (! sc->vptr->is_number (pair_car (sink)))
+      return ffi_sprintf (sc, "%d%s argument is not a number",
+                          ffi_arg_index, ordinal_suffix (ffi_arg_index));
 
-      bytes_read = read (source, buffer, want);
+  while (1)
+    {
+      bytes_read = read (source, buffer, sizeof buffer);
       if (bytes_read == 0)
         break;
       if (bytes_read < 0)
         FFI_RETURN_ERR (sc, gpg_error_from_syserror ());
-      if (write (sink, buffer, bytes_read) != bytes_read)
-        FFI_RETURN_ERR (sc, gpg_error_from_syserror ());
-      if (len != -1)
-        len -= bytes_read;
+
+      for (sink = sinks; sink != sc->NIL; sink = pair_cdr (sink))
+        {
+          int fd = sc->vptr->ivalue (pair_car (sink));
+          char *p = buffer;
+          ssize_t left = bytes_read;
+
+          while (left)
+            {
+              ssize_t written = write (fd, p, left);
+              if (written < 0)
+                FFI_RETURN_ERR (sc, gpg_error_from_syserror ());
+              assert (written <= left);
+              left -= written;
+              p += written;
+            }
+        }
     }
   FFI_RETURN (sc);
 }
@@ -1015,42 +1137,6 @@ do_string_contains (scheme *sc, pointer args)
   FFI_RETURN_POINTER (sc, strstr (haystack, needle) ? sc->T : sc->F);
 }
 
-static pointer
-do_glob (scheme *sc, pointer args)
-{
-  FFI_PROLOG ();
-  pointer result = sc->NIL;
-  size_t i;
-  char *pattern;
-  glob_t pglob;
-  FFI_ARG_OR_RETURN (sc, char *, pattern, string, args);
-  FFI_ARGS_DONE_OR_RETURN (sc, args);
-
-  switch (glob (pattern, 0, NULL, &pglob))
-    {
-    case 0:
-      for (i = 0; i < pglob.gl_pathc; i++)
-        result =
-          (sc->vptr->cons) (sc,
-                            sc->vptr->mk_string (sc, pglob.gl_pathv[i]),
-                            result);
-      globfree (&pglob);
-      break;
-
-    case GLOB_NOMATCH:
-      /* Return the empty list.  */
-      break;
-
-    case GLOB_NOSPACE:
-      return ffi_sprintf (sc, "out of memory");
-    case GLOB_ABORTED:
-      return ffi_sprintf (sc, "read error");
-    default:
-      assert (! "not reached");
-    }
-  FFI_RETURN_POINTER (sc, result);
-}
-
 
 
 static pointer
@@ -1209,7 +1295,8 @@ ffi_scheme_eval (scheme *sc, const char *format, ...)
 }
 
 gpg_error_t
-ffi_init (scheme *sc, const char *argv0, int argc, const char **argv)
+ffi_init (scheme *sc, const char *argv0, const char *scriptname,
+          int argc, const char **argv)
 {
   int i;
   pointer args = sc->NIL;
@@ -1237,6 +1324,9 @@ ffi_init (scheme *sc, const char *argv0, int argc, const char **argv)
   ffi_define_constant (sc, STDIN_FILENO);
   ffi_define_constant (sc, STDOUT_FILENO);
   ffi_define_constant (sc, STDERR_FILENO);
+  ffi_define_constant (sc, SEEK_SET);
+  ffi_define_constant (sc, SEEK_CUR);
+  ffi_define_constant (sc, SEEK_END);
 
   ffi_define_function (sc, sleep);
   ffi_define_function (sc, usleep);
@@ -1244,10 +1334,11 @@ ffi_init (scheme *sc, const char *argv0, int argc, const char **argv)
   ffi_define_function (sc, strerror);
   ffi_define_function (sc, getenv);
   ffi_define_function (sc, setenv);
-  ffi_define_function (sc, exit);
+  ffi_define_function_name (sc, "_exit", exit);
   ffi_define_function (sc, open);
   ffi_define_function (sc, fdopen);
   ffi_define_function (sc, close);
+  ffi_define_function (sc, seek);
   ffi_define_function_name (sc, "_mkdtemp", mkdtemp);
   ffi_define_function (sc, unlink);
   ffi_define_function (sc, unlink_recursively);
@@ -1255,6 +1346,13 @@ ffi_init (scheme *sc, const char *argv0, int argc, const char **argv)
   ffi_define_function (sc, getcwd);
   ffi_define_function (sc, mkdir);
   ffi_define_function (sc, rmdir);
+  ffi_define_function (sc, get_isotime);
+  ffi_define_function (sc, getpid);
+
+  /* Random numbers.  */
+  ffi_define_function (sc, srandom);
+  ffi_define_function (sc, random);
+  ffi_define_function (sc, make_random_string);
 
   /* Process management.  */
   ffi_define_function (sc, spawn_process);
@@ -1277,7 +1375,6 @@ ffi_init (scheme *sc, const char *argv0, int argc, const char **argv)
   ffi_define_function (sc, string_index);
   ffi_define_function (sc, string_rindex);
   ffi_define_function_name (sc, "string-contains?", string_contains);
-  ffi_define_function (sc, glob);
 
   /* User interface.  */
   ffi_define_function (sc, flush_stdio);
@@ -1288,6 +1385,7 @@ ffi_init (scheme *sc, const char *argv0, int argc, const char **argv)
   ffi_define_function_name (sc, "*set-verbose!*", set_verbose);
 
   ffi_define (sc, "*argv0*", sc->vptr->mk_string (sc, argv0));
+  ffi_define (sc, "*scriptname*", sc->vptr->mk_string (sc, scriptname));
   for (i = argc - 1; i >= 0; i--)
     {
       pointer value = sc->vptr->mk_string (sc, argv[i]);
@@ -1301,6 +1399,15 @@ ffi_init (scheme *sc, const char *argv0, int argc, const char **argv)
   ffi_define (sc, "*pathsep*", sc->vptr->mk_character (sc, ':'));
 #endif
 
+  ffi_define (sc, "*win32*",
+#if _WIN32
+              sc->T
+#else
+              sc->F
+#endif
+              );
+
+
   ffi_define (sc, "*stdin*",
               sc->vptr->mk_port_from_file (sc, stdin, port_input));
   ffi_define (sc, "*stdout*",
